package com.nisum.ecomcustomer.service.impl;

import com.nisum.ecomcustomer.exceptions.CustomerNotFoundException;
import com.nisum.ecomcustomer.exceptions.EmailAlreadyRegisteredException;
import com.nisum.ecomcustomer.exceptions.InvalidDateRangeException;
import com.nisum.ecomcustomer.model.Customer;
import com.nisum.ecomcustomer.model.CustomerType;
import com.nisum.ecomcustomer.repository.CustomerRepository;
import com.nisum.ecomcustomer.service.CustomerService;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.data.mongodb.core.MongoTemplate;
import org.springframework.data.mongodb.core.query.Query;
import org.springframework.stereotype.Service;

import java.time.LocalDate;
import java.util.List;
import java.util.stream.Collectors;

import static org.springframework.data.mongodb.core.query.Criteria.where;

@Slf4j
@Service
public class CustomerServiceImpl implements CustomerService {

    @Autowired
    CustomerRepository customerRepository;

    @Autowired
    MongoTemplate mongoTemplate;

    @Override
    public Customer register(Customer customer) {
        customer.setId(null); // auto-generated
        if (customerRepository.existsByEmail(customer.getEmail())) {
            throw new EmailAlreadyRegisteredException();
        }
        return customerRepository.save(customer);
    }

    @Override
    public List<Customer> registerAll(List<Customer> customers) {
        List<String> emails = customers.stream().map(Customer::getEmail).collect(Collectors.toList());
        if (customerRepository.existsByEmailIn(emails))
            throw new EmailAlreadyRegisteredException("one of the emails is already registered");
        return customerRepository.saveAll(customers);
    }

    @Override
    public Customer update(Customer customer) {
        Customer existingCustomer;
        if (customer.getId() != null) {
            existingCustomer = customerRepository.findById(customer.getId()).orElseThrow(CustomerNotFoundException::new);
            //createdDate is prone to be changed by client
            customer.setCreatedDate(existingCustomer.getCreatedDate());
        } else
            throw new CustomerNotFoundException();
        return customerRepository.save(customer);
    }

    @Override
    public List<Customer> getAllCustomers() {
        return customerRepository.findAll();
    }

    @Override
    public Customer getCustomerById(Long id) {
        return customerRepository.findById(id).orElseThrow(CustomerNotFoundException::new);
    }

    @Override
    public Customer getCustomerByEmail(String email) {
        return customerRepository.findByEmailIgnoreCase(email).orElseThrow(CustomerNotFoundException::new);
    }

    @Override
    public List<Customer> getCustomerByFirstName(String fName) {
        return customerRepository.findByFirstName(fName);
    }

    @Override
    public List<Customer> getCustomerByLastName(String lName) {
        return customerRepository.findByLastNameIgnoreCase(lName);
    }

    @Override
    public List<Customer> getCustomersByCustomerType(CustomerType customerType) {
        return customerRepository.findByCustomerType(customerType);
    }

    @Override
    public List<Customer> getRegisteredCustomersOnDate(LocalDate onDate) {
        return customerRepository.findByCreatedDate(onDate);
    }

    @Override
    public List<Customer> getRegisteredCustomersBetweenDates(LocalDate fromDate, LocalDate toDate) {
        if (fromDate.isAfter(toDate)) throw new InvalidDateRangeException("invalid date range: fromDate can't be greater than toDate");
        Query query = new Query(where("createdDate").gte(fromDate).lte(toDate));
        return mongoTemplate.find(query, Customer.class);
    }

    @Override
    public void deleteById(Long id) {
        if (customerRepository.existsById(id)) customerRepository.deleteById(id);
        else throw new CustomerNotFoundException(id);
    }

}
