package com.nisum.ecomcustomer.controller;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nisum.ecomcustomer.exceptions.CustomerNotFoundException;
import com.nisum.ecomcustomer.exceptions.CustomerServiceExceptionHandler;
import com.nisum.ecomcustomer.exceptions.EmailAlreadyRegisteredException;
import com.nisum.ecomcustomer.model.Address;
import com.nisum.ecomcustomer.model.AddressType;
import com.nisum.ecomcustomer.model.Customer;
import com.nisum.ecomcustomer.model.CustomerType;
import com.nisum.ecomcustomer.service.CustomerService;
import org.junit.jupiter.api.Test;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.WebMvcTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.http.MediaType;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.MvcResult;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.util.Arrays;
import java.util.List;
import java.util.stream.Collectors;

import static org.hamcrest.Matchers.hasSize;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.*;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.jsonPath;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

@WebMvcTest(value = {CustomerController.class, CustomerServiceExceptionHandler.class})
class CustomerControllerTest {

    @MockBean
    CustomerService customerService;

    @Autowired
    MockMvc mockMvc;

    @Autowired
    ObjectMapper objectMapper;

    @Test
    void registerCustomer_returnsCustomer_201() throws Exception {
        Customer customerRequestObj = getCustomerRequestObject();
        Customer customerResponseObj = getCustomerResponseObject();
        when(customerService.register(customerRequestObj)).thenReturn(customerResponseObj);
        MvcResult result = mockMvc.perform(post("/customer/register")
                .content(objectMapper.writeValueAsString(customerRequestObj))
                .contentType(MediaType.APPLICATION_JSON))
                .andExpect(status().isCreated())
                .andReturn();
        Customer response = getObjectFromResponse(result, Customer.class);
        assertEquals("Ashok", response.getFirstName());
        assertEquals(3, response.getAddresses().size());
        assertNotNull(response.getId());
        assertEquals(LocalDate.now(), response.getCreatedDate());
    }

    @Test
    void registerCustomer_throwsEmailAlreadyRegisteredException_409() throws Exception {
        Customer customerRequestObject = getCustomerRequestObject();
        when(customerService.register(customerRequestObject)).thenThrow(new EmailAlreadyRegisteredException());
        mockMvc.perform(post("/customer/register")
                .content(objectMapper.writeValueAsString(customerRequestObject))
                .contentType(MediaType.APPLICATION_JSON))
                .andExpect(status().isConflict())
                .andReturn();
    }

    @Test
    void registerCustomers_returnsCustomers_201() throws Exception {
        List<Customer> customersRequestObj = getCustomerRequestObjectsList();
        List<Customer> customersResponseObj = getCustomerResponseObjectsList();
        when(customerService.registerAll(customersRequestObj)).thenReturn(customersResponseObj);
        mockMvc.perform(post("/customer/registerAll")
                .content(objectMapper.writeValueAsString(customersRequestObj))
                .contentType(MediaType.APPLICATION_JSON))
                .andExpect(status().isCreated())
                .andExpect(jsonPath("$", hasSize(3)))
                .andReturn();
    }

    @Test
    void registerCustomers_throwsEmailAlreadyRegisteredException_409() throws Exception {
        List<Customer> customersRequestObj = getCustomerRequestObjectsList();
        when(customerService.registerAll(customersRequestObj)).thenThrow(new EmailAlreadyRegisteredException());
        mockMvc.perform(post("/customer/registerAll")
                .content(objectMapper.writeValueAsString(customersRequestObj))
                .contentType(MediaType.APPLICATION_JSON))
                .andExpect(status().isConflict())
                .andReturn();
    }

    @Test
    void updateCustomer_returnsCustomer_201() throws Exception {
        Customer customerRequestObj = getCustomerRequestObject();
        customerRequestObj.setId(1L);
        customerRequestObj.setEmail(customerRequestObj.getEmail().toLowerCase());
        customerRequestObj.setCreatedDate(LocalDate.now());
        customerRequestObj.setModifiedDate(LocalDateTime.now());
        Customer customerResponseObj = getCustomerResponseObject();
        when(customerService.update(customerRequestObj)).thenReturn(customerResponseObj);
        MvcResult result = mockMvc.perform(put("/customer/update")
                .content(objectMapper.writeValueAsString(customerRequestObj))
                .contentType(MediaType.APPLICATION_JSON))
                .andExpect(status().isCreated())
                .andReturn();
        Customer response = getObjectFromResponse(result, Customer.class);
        assertEquals("Ashok", response.getFirstName());
        assertEquals(3, response.getAddresses().size());
        assertNotNull(response.getId());
        assertEquals(LocalDate.now(), response.getCreatedDate());
        assertNotEquals(customerRequestObj.getModifiedDate(), response.getModifiedDate());
    }

    @Test
    void updateCustomer_throwsCustomerNotFoundException_404() throws Exception {
        Customer customerRequestObj = getCustomerRequestObject();
        customerRequestObj.setId(null);
        when(customerService.update(customerRequestObj)).thenThrow(new CustomerNotFoundException());
        mockMvc.perform(put("/customer/update")
                .content(objectMapper.writeValueAsString(customerRequestObj))
                .contentType(MediaType.APPLICATION_JSON))
                .andExpect(status().isNotFound())
                .andReturn();
    }

    @Test
    void getAllCustomers_returnsCustomers_200() throws Exception {
        when(customerService.getAllCustomers()).thenReturn(getCustomerResponseObjectsList());
        mockMvc.perform(get("/customer/"))
                .andExpect(status().isOk())
                .andExpect(jsonPath("$", hasSize(3)))
                .andReturn();
    }

    @Test
    void getCustomerById_returnsCustomer_200() throws Exception {
        when(customerService.getCustomerById(1L)).thenReturn(getCustomerResponseObject());
        mockMvc.perform(get("/customer/id/" + 1))
                .andExpect(status().isOk())
                .andExpect(jsonPath("$.id", is(1)))
                .andReturn();
    }

    @Test
    void getCustomerById_throwsCustomerNotFoundException_404() throws Exception {
        when(customerService.getCustomerById(1L)).thenThrow(new CustomerNotFoundException());
        mockMvc.perform(get("/customer/id/" + 1))
                .andExpect(status().isNotFound())
                .andReturn();
    }

    @Test
    void getCustomerByEmail_returnsCustomer_200() throws Exception {
        String email = "ashokk@nisum.com";
        when(customerService.getCustomerByEmail(email)).thenReturn(getCustomerResponseObjectsList().stream().filter(customer -> customer.getEmail().equals(email)).findFirst().get());
        mockMvc.perform(get("/customer/email/" + email))
                .andExpect(status().isOk())
                .andExpect(jsonPath("$.email", is(email)))
                .andReturn();
    }

    @Test
    void getCustomerByEmail_throwsCustomerNotFoundException_404() throws Exception {
        String email = "ashokk@nisum.com";
        when(customerService.getCustomerByEmail(email)).thenThrow(new CustomerNotFoundException());
        mockMvc.perform(get("/customer/email/" + email))
                .andExpect(status().isNotFound())
                .andReturn();
    }

    @Test
    void getCustomersByFirstName() {
    }

    @Test
    void getCustomersByLastName() {
    }


    @Test
    void getCustomersByCustomerType_returnsCustomers_200() throws Exception {
        int loyalValue = CustomerType.LOYAL.getValue();
        when(customerService.getCustomersByCustomerType(CustomerType.LOYAL)).
                thenReturn(getCustomerResponseObjectsList().stream()
                        .filter(customer -> customer.getCustomerType().getValue() == loyalValue)
                        .collect(Collectors.toList()));
        mockMvc.perform(get("/customer/type/" + loyalValue))
                .andExpect(status().isOk())
                .andExpect(jsonPath("$", hasSize(2)))
                .andExpect(jsonPath("$[0].customerType", is(loyalValue)))
                .andExpect(jsonPath("$[1].customerType", is(loyalValue)))
                .andReturn();
    }

    @Test
    void getCustomersByCustomerType_throwsIllegalArgumentException_400() throws Exception {
        int typeValue = 3;
        mockMvc.perform(get("/customer/type/" + typeValue))
                .andExpect(status().isBadRequest())
                .andReturn();
    }

    @Test
    void getRegisteredCustomersOnDate() {
    }

    @Test
    void getRegisteredCustomersBetweenDates() {
    }

    @Test
    void deleteById_returnsNoContent_204() throws Exception {
        doNothing().when(customerService).deleteById(1L);
        mockMvc.perform(delete("/customer/id/" + 1))
                .andExpect(status().isNoContent());
    }

    @Test
    void deleteById_throwsCustomerNotFoundException_404() throws Exception {
        doThrow(new CustomerNotFoundException()).when(customerService).deleteById(1L);
        mockMvc.perform(delete("/customer/id/" + 1))
                .andExpect(status().isNotFound());
    }

    private Customer getCustomerRequestObject() {
        Customer customer = new Customer();
        customer.setId(null);
        customer.setFirstName("Ashok");
        customer.setLastName("Kumar");
        customer.setEmail("ashokk@nisum.com");
        customer.setDayPhone("8989898989");
        customer.setCustomerType(CustomerType.LOYAL);


        Address billingAddress = new Address();
        billingAddress.setAddressLine1("Srinagar colony");
        billingAddress.setAddressLine2("Ameerpet");
        billingAddress.setAddressType(AddressType.BILLING);
        billingAddress.setDefault(true);
        billingAddress.setZipCode("500000");
        billingAddress.setCity("Hyderabad");
        billingAddress.setState("TS");
        billingAddress.setCountry("IND");

        Address shippingAddress1 = new Address();
        shippingAddress1.setAddressLine1("Jawaharnagar colony");
        shippingAddress1.setAddressLine2("Ameerpet");
        shippingAddress1.setAddressType(AddressType.SHIPPING);
        shippingAddress1.setDefault(true);
        shippingAddress1.setZipCode("500001");
        shippingAddress1.setCity("Hyderabad");
        shippingAddress1.setState("TS");
        shippingAddress1.setCountry("IND");

        Address shippingAddress2 = new Address();
        shippingAddress2.setAddressLine1("Krishnanagar colony");
        shippingAddress2.setAddressLine2("Ameerpet");
        shippingAddress2.setAddressType(AddressType.SHIPPING);
        shippingAddress2.setDefault(false);
        shippingAddress2.setZipCode("500002");
        shippingAddress2.setCity("Hyderabad");
        shippingAddress2.setState("TS");
        shippingAddress2.setCountry("IND");

        List<Address> addresses = Arrays.asList(billingAddress, shippingAddress1, shippingAddress2);
        customer.setAddresses(addresses);

        return customer;
    }

    private Customer getCustomerResponseObject() {
        Customer customer = getCustomerRequestObject();
        customer.setId(1L);
        customer.setEmail(customer.getEmail().toLowerCase());
        customer.setCreatedDate(LocalDate.now());
        customer.setModifiedDate(LocalDateTime.now().plusSeconds(2));
        return customer;
    }

    private List<Customer> getCustomerRequestObjectsList() {
        Customer customer1 = getCustomerRequestObject();
        Customer customer2 = getCustomerRequestObject();
        customer2.setEmail("anandk@nisum.com");
        Customer customer3 = getCustomerRequestObject();
        customer3.setEmail("anandk1@nisum.com");
        customer3.setCustomerType(CustomerType.REGULAR);
        return Arrays.asList(customer1, customer2, customer3);
    }

    private List<Customer> getCustomerResponseObjectsList() {
        List<Customer> customers = getCustomerRequestObjectsList();
        Long id = 1L;
        for (Customer customer : customers) {
            customer.setId(id++);
            customer.setEmail(customer.getEmail().toLowerCase());
            customer.setCreatedDate(LocalDate.now());
            customer.setModifiedDate(LocalDateTime.now());
        }
        return customers;
    }

    private <T> T getObjectFromResponse(MvcResult result, Class<T> targetClass) throws Exception {
        String responseAsString = result.getResponse().getContentAsString();
        return objectMapper.readValue(responseAsString, targetClass);
    }

}